

#!/usr/bin/python2
# coding=utf-8
# Copyright (c) Microsoft. All rights reserved.

# Helper python script with utility methods

import sys
import os
import grp
import pwd
import getpass
import unicodedata
import subprocess
import platform
import gettext
import locale
import re
from ConfigParser import ConfigParser
import socket

#
# Static configuration values
#
sqlPathRoot = "/var/opt/mssql"
configurationFilePath = os.path.join(sqlPathRoot, "mssql.conf")
masterDatabaseFilePath = os.path.join(sqlPathRoot, "data", "master.mdf")
eulaConfigSection="EULA"
eulaConfigSetting="accepteula"
eulaMlConfigSetting="accepteulaml"
telemetryConfigSetting="customerfeedback"
telemetryLocalAuditCacheDirectorySetting="userrequestedlocalauditdirectory"
errorExitCode = 1
successExitCode = 0
directoryOfScript = os.path.dirname(os.path.realpath(__file__))
checkInstallScript = directoryOfScript + "/checkinstall.sh"
sqlBinPathRoot = "/opt/mssql/bin"
launchpaddPath = sqlBinPathRoot + "/launchpadd"
mlservicesPathRoot = "/opt/mssql/mlservices"
checkInstallExtensibilityScript = sqlBinPathRoot + "/checkinstallextensibility.sh"
checkRunningInstanceScript = directoryOfScript + "/checkrunninginstance.sh"
invokeSqlservrScript = directoryOfScript + "/invokesqlservr.sh"
setCollationScript = directoryOfScript + "/set-collation.sh"
sudo = "sudo"
mssqlUser = "mssql"
saPasswordEnvVariable = "SA_PASSWORD"
mssqlSaPasswordEnvVariable = "MSSQL_SA_PASSWORD"
mssqlLcidEnvVariable = "MSSQL_LCID"
mssqlPidEnvVariable = "MSSQL_PID"
language = "language"
lcid = "lcid"
expressEdition = "express"
evaluationEdition = "evaluation"
developerEdition = "developer"
webEdition = "web"
standardEdition = "standard"
enterpriseEdition = "enterprise"
enterpriseCoreEdition = "enterprisecore"
supportedLcids = ['1033', '1031', '3082', '1036', '1040',
                  '1041', '1042', '1046', '1049', '2052', '1028']
spnPrefix = "MSSQLSvc"
spnDefaultPort = "1433"
kvnoCommandFormat = "kvno %s"
addKeytabEntryCommandFormat = """echo \"addent -password -p %s -k %s -e %s 
%s
wkt %s
quit
\" | ktutil"""
keytabEncryptionTypes = ["aes256-cts-hmac-sha1-96", "arcfour-hmac"]
listKeytabContentsCommandFormat = "klist -kte %s"
copyKeytabEntriesPrefixCommandFormat = "echo \"rkt %s\n"
deleteKeytabEntryCommandFormat = "delent %s\n"
copyKeytabEntriesSuffixCommandFormat = """wkt %s
quit
\" | ktutil"""
rootKeytab = "/etc/krb5.keytab"

#
# Colors to use when printing to standard out
#
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

#
# Error codes for failed password validation
#
class passwordErrorCodes:
    SUCCESS        = 0
    TOO_SHORT      = 1
    TOO_LONG       = 2
    NOT_COMPLEX    = 3
    ENCODING_ERROR = 4
    DECODING_ERROR = 5
    CONTROL_CHARS  = 6

def printError(text):
    """printError

    Args:
        text(str): Text to print
    """

    _printTextInColor(text, bcolors.RED)

def printErrorUnsupportedSetting(section_name, setting_name=None):
    """Print error message and exit for unsupported settings

    Args:
        section_name(str): The section name
                           For legacy reasons section_name can contain setting_name as well
                           For example: section_name.setting_name
        setting_name(str): The setting name
    """

    if setting_name:
        print (_("The section '%s' and setting '%s' is not supported.") % (section_name, setting_name))
    else:
        print(_("The setting '%s' is not supported.") % section_name)
    exit(errorExitCode)

def checkColorSupported():
    """Check if color is supported

    Returns:
        True if color is supported, False otherwise
    """

    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or 'ANSICON' in os.environ)
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()

    if not supported_platform or not is_a_tty:
        return False

    return True

def setSettingsFromEnvironment():
    """Set settings from environment
    """

    import mssqlsettingsmanager

    ret = True

    for setting in mssqlsettingsmanager.supportedSettingsList:
        if setting.environment_variable in os.environ:
            # Make a local copy of environment variable and then remove
            # it from the environment. Otherwise it will be passed on
            # to SQL Server when running setup.
            #
            value = os.environ[setting.environment_variable]
            del os.environ[setting.environment_variable]

            ret = setSetting(setting, value)

            if not ret:
                break

    return ret

def setSettingByName(section, name, value):
    """ Set a value in mssql.conf
    """

    import mssqlsettingsmanager

    setting = mssqlsettingsmanager.findSetting(section, name)

    if setting == None:
        print(_("Warning: Failed to set '%s.%s' to '%s' in mssql.conf. Please set this manually") % (section, name, value))
        return False

    return setSetting(setting, value)


def setSetting(setting, value):
    """ Set a value in mssql.conf
    """

    import mssqlsettingsmanager

    config = ConfigParser()
    readConfigFromFile(config, configurationFilePath)

    if len(value) > 0:
        if setting.sectionOnly:
            if mssqlsettingsmanager.setSectionOnlySettings(config, setting, value, True):
                writeConfigToFile(config, configurationFilePath)
            else:
                return False
        elif mssqlsettingsmanager.setSetting(config, setting, value, True):
                writeConfigToFile(config, configurationFilePath)
        else:
            return False
    else:
        if setting.sectionOnly:
            if mssqlsettingsmanager.unsetSectionOnlySettings(config, setting):
                writeConfigToFile(config, configurationFilePath)
        elif mssqlsettingsmanager.unsetSetting(config, setting):
                writeConfigToFile(config, configurationFilePath)

    return True

def languageSelect(noprompt):
    """Select language

    Args:
        noprompt(boolean): True if --noprompt specified, false otherwise
    """

    lcidFromEnv = os.environ.get(mssqlLcidEnvVariable)

    if (lcidFromEnv != None):
        print _("Setting language using LCID from environment variable %s") % mssqlLcidEnvVariable
        writeLcidToConfFile(lcidFromEnv)
        return

    if(noprompt == False):
        language = locale.getdefaultlocale()[0]
        if(language == None or language == "" or language.lower() == "en_us"):
            # Nothing to do as en_US will be chosen by default by the engine
            return
        else:
            print ""
            print _("Choose the language for SQL Server:")
            print (u"(1) English")
            print (u"(2) Deutsch")
            print (u"(3) Español")
            print (u"(4) Français")
            print (u"(5) Italiano")
            print (u"(6) 日本語")
            print (u"(7) 한국어")
            print (u"(8) Português")
            print (u"(9) Русский")
            print (u"(10) 中文 – 简体")
            print (u"(11) 中文 （繁体）")

            languageOption = raw_input(_("Enter Option 1-11: "))

            optionToLcid = { '1': '1033', #en-US
                     '2': '1031', #de-DE
                     '3': '3082', #es-ES
                     '4': '1036', #fr-FR
                     '5': '1040', #it-IT
                     '6': '1041', #ja-JP
                     '7': '1042', #ko-KR
                     '8': '1046', #pt-BR
                     '9': '1049', #ru-RU
                     '10': '2052', #zh-CN
                     '11': '1028'} #zh-TW

            if (languageOption in optionToLcid.keys()):
                writeLcidToConfFile(optionToLcid[languageOption])
            else:
                print _("Invalid Option. Exiting.")
                exit(errorExitCode)


def isValidLcid(lcidValue):
    """Check if a LCID value is valid.

    Args:
        lcidValue(int): LCID value
    """

    return lcidValue in supportedLcids

def writeLcidToConfFile(lcidValue):
    """Write LCID to configuration file

    Args:
        lcidValue(int): LCID value
    """

    if (isValidLcid(lcidValue) == False):
        print _("LCID %s is not supported.") % lcidValue
        exit(errorExitCode)

    config = ConfigParser(allow_no_value=True)
    readConfigFromFile(config, configurationFilePath)

    if (config.has_section(language) == False):
        config.add_section(language)

    config.set(language, lcid, lcidValue)
    writeConfigToFile(config, configurationFilePath)

def printEngineWelcomeMessage():
    """Print engine welcome message
    """

    print ("")
    print ("+--------------------------------------------------------------+")
    print (_("Please run 'sudo /opt/mssql/bin/mssql-conf setup'"))
    print (_("to complete the setup of Microsoft SQL Server"))
    print ("+--------------------------------------------------------------+")
    print ("")
    return successExitCode

def printAgentWelcomeMessage():
    """Print agent welcome message
    """

    print ("")
    print ("+--------------------------------------------------------------------------------+")
    print (_("Please restart mssql-server to enable Microsoft SQL Server Agent."))
    print ("+--------------------------------------------------------------------------------+")
    print ("")

def printFTSWelcomeMessage():
    """Print full-text welcome message
    """

    print ("")
    print ("+-------------------------------------------------------------------------------------+")
    print (_("Please restart mssql-server to enable Microsoft SQL Server Full Text Search."))
    print ("+-------------------------------------------------------------------------------------+")
    print ("")

def printPolyBaseWelcomeMessage():
    """Print PolyBase welcome message
    """

    print ("")
    print ("+--------------------------------------------------------------------------------+")
    print (_("Please restart mssql-server to enable PolyBase."))
    print ("+--------------------------------------------------------------------------------+")
    print ("")

def printPolyBaseHadoopWelcomeMessage():
    """Print PolyBase Hadoop welcome message
    """

    print ("")
    print ("+--------------------------------------------------------------------------------+")
    print (_("Please restart mssql-launchpadd to enable PolyBase Hadoop."))
    print ("+--------------------------------------------------------------------------------+")
    print ("")

def printMachineLearningServicesWelcomeMessage():
    """Print the Machine Learning services welcome message
    """

    print ("")
    print ("+-------------------------------------------------------------------------------------+")
    print (_("Please run 'sudo /opt/mssql/bin/mssql-conf setup accept-eula-ml'"))
    print (_("to accept the Machine Learning Services EULA."))
    print ("+-------------------------------------------------------------------------------------+")
    print ("")

def getFwlinkWithLocale(linkId):
    """Gets the correct Url for the fwlink based on the users locale

    Args:
        linkId(string): The fwlink ID

    Returns:
        The string with the complete url
    """

    baseUrl = "https://go.microsoft.com/fwlink/?LinkId=" + linkId
    localeCode = locale.getlocale()[0]
    localeToClcid = {'en_US': '0x409',  # en-US
                     'de_DE': '0x407',  # de-DE
                     'es_ES': '0x40a',  # es-ES
                     'fr_FR': '0x40c',  # fr-FR
                     'it_IT': '0x410',  # it-IT
                     'ja_JP': '0x411',  # ja-JP
                     'ko_KR': '0x412',  # ko-KR
                     'pt_BR': '0x416',  # pt-BR
                     'ru_RU': '0x419',  # ru-RU
                     'zh_CN': '0x804',  # zh-CN
                     'zh_TW': '0x404'}  # zh-TW

    if localeCode in localeToClcid:
        return baseUrl + "&clcid=" + localeToClcid[localeCode]
    else:
        return baseUrl


def checkEulaAgreement(eulaAccepted, configurationFilePath, ignoreMasterDatabase=False, isEvaluationEdition = False):
    """Check if the EULA agreement has been accepted.

    Args:
        eulaAccepted(boolean): User has indicated their acceptance via command-line
                               or environment variable.
        configurationFilePath(str): Configuration file path
        ignoreMasterDatabase(boolean): Ignore presence of master database
        isEvaluationEdition(boolean): True if edition selected is evaluation, false otherwise

    Returns:
        True if accepted, False otherwise
    """

    print(_("The license terms for this product can be found in"))
    print("/usr/share/doc/mssql-server " + _("or downloaded from:"))
    if isEvaluationEdition:
        print(getFwlinkWithLocale("2104078"))
    else:
        print(getFwlinkWithLocale("2104294"))
    print("")
    print(_("The privacy statement can be viewed at:"))
    print(getFwlinkWithLocale("853010"))
    print("")

    if os.path.exists(masterDatabaseFilePath) and not ignoreMasterDatabase:
        return True

    config = ConfigParser(allow_no_value=True)
    readConfigFromFile(config, configurationFilePath)

    if (config.has_section(eulaConfigSection) == False or \
        config.get(eulaConfigSection, eulaConfigSetting) is None):
        if not eulaAccepted:
            agreement = raw_input(_("Do you accept the license terms?") + " [Yes/No]:")
            print("")

            if (agreement.strip().lower() == "yes" or agreement.strip().lower() == "y"):
                eulaAccepted = True
            else:
                return False

        if eulaAccepted:
            config.add_section(eulaConfigSection)
            config.set(eulaConfigSection, eulaConfigSetting, "Y")
            writeConfigToFile(config, configurationFilePath)
            return True

    return True

def checkEulaMlAgreement(eulaMlAccepted, configurationFilePath):
    """Check if the EULA agreement for machine learning services has been accepted.

    Args:
        eulaMlAccepted(boolean): User has indicated their acceptance via command-line
                               or environment variable.
        configurationFilePath(str): Configuration file path

    Returns:
        True if accepted, False otherwise
    """

    print(_("The license terms for this product can be downloaded from:"))
    print(getFwlinkWithLocale("2006040"))

    config = ConfigParser(allow_no_value=True)
    readConfigFromFile(config, configurationFilePath)

    if (config.has_section(eulaConfigSection) == True and \
        (config.has_option(eulaConfigSection, eulaConfigSetting) and config.get(eulaConfigSection, eulaConfigSetting) == "Y")):
        if (config.has_option(eulaConfigSection, eulaMlConfigSetting) and config.get(eulaConfigSection, eulaMlConfigSetting) == "Y"):
            return True
        else:
            if not eulaMlAccepted:
                agreement = raw_input(_("Do you accept the license terms for machine learning services?") + " [Yes/No]:")
                print("")

                if (agreement.strip().lower() == "yes" or agreement.strip().lower() == "y"):
                    eulaMlAccepted = True
                else:
                    return False

            if eulaMlAccepted:
                config.set(eulaConfigSection, eulaMlConfigSetting, "Y")
                writeConfigToFile(config, configurationFilePath)
                return True

    else:
        print (_("The SQL Server EULA needs to be accepted first"))
        print ("")
        return False

def checkSudo():
    """Check if we're running as root

    Returns:
        True if running as root, False otherwise
    """

    if (os.geteuid() == 0):
        return True

    return False

def checkSudoOrMssql():
    """Check if we're running as root or the user is in the mssql group.

    Returns:
        True if running as root or in mssql group, False otherwise
    """

    if(checkSudo() == True):
        return True

    user = getpass.getuser()
    groups = [g.gr_name for g in grp.getgrall() if user in g.gr_mem]
    gid = pwd.getpwnam(user).pw_gid
    groups.append(grp.getgrgid(gid).gr_name)

    if(('mssql' in groups) and (user == mssqlUser)):
        return True

    return False

def printValidationErrorMessage(setting, errorMessage):
    """Print a validation error message.

    Args:
        setting(str): Setting name
        errorMessage(str): Error message
    """

    printError(_("Validation error on setting '%s.%s'") % (setting.section, setting.name))
    printError(errorMessage)

def printPasswordErrorMessage(errorCode):
    """Print an error message if password can't be validated.

     Args:
         errorCode(int): Error code
    """

    if errorCode == passwordErrorCodes.TOO_SHORT:
        printError((_("The specified password does not meet SQL Server password policy requirements because it is "
                "too short. The password must be at least 8 characters")))
    elif errorCode == passwordErrorCodes.TOO_LONG:
        printError((_("The specified password does not meet SQL Server password policy requirements because it is "
                "too long. The password cannot exceed 128 characters")))
    elif errorCode == passwordErrorCodes.NOT_COMPLEX:
        printError((_("The specified password does not meet SQL Server password policy requirements because it "
                      "is not complex enough. The password must be at least 8 characters long and contain characters "
                      "from three of the following four sets: uppercase letters, lowercase letters, numbers, "
                      "and symbols.")))
    elif errorCode == passwordErrorCodes.ENCODING_ERROR:
        printError(_("The specified password contains a character that cannot be encoded to UTF-8. "
                         "Try using a password with only ASCII characters."))
    elif errorCode == passwordErrorCodes.DECODING_ERROR:
        printError(_("The specified password contains a character that cannot be decoded. "
                         "Try using a password with only ASCII characters."))
    elif errorCode == passwordErrorCodes.CONTROL_CHARS:
        printError((_("The specified password contains an invalid character. Valid characters "
                               "include uppercase letters, lowercase letters, numbers, symbols, "
                               "punctuation marks, and unicode characters that are categorized as alphabetic "
                               "but are not uppercase or lowercase.")))

def makeDirectoryIfNotExists(directoryPath):
    """Make a directory if it does not exist

    Args:
        directoryPath(str): Directory path
    """

    try:
        if os.path.exists(directoryPath):
            return
        if not os.path.exists(os.path.dirname(directoryPath)):
            makeDirectoryIfNotExists(os.path.dirname(directoryPath))
        os.makedirs(directoryPath)
    except IOError, err:
        if err.errno == 13:
            printError(_("Permission denied to mkdir '%s'.") % (directoryPath))
            exit(errorExitCode)
        else:
            printError(err)
            exit(errorExitCode)

def writeConfigToFile(config, configurationFilePath):
    """Write configuration to a file

    Args:
        config(object): Config parser object
        configurationFilePath(str): Configuration file path
    """

    makeDirectoryIfNotExists(os.path.dirname(configurationFilePath))

    try:
        with open(configurationFilePath, 'w') as configFile:
            config.write(configFile)
    except IOError, err:
        if err.errno == 13:
            printError(_("Permission denied to modify SQL Server configuration."))
        else:
            printError(err)
            exit(errorExitCode)

def readConfigFromFile(config, configurationFilePath):
    """"Read configuration from a file

    Args:
        config(object): Config parser object
        configurationFilePath(str): Configuration file path
    """

    if (os.path.exists(configurationFilePath) == True):
        try:
            config.read(configurationFilePath)
        except:
            printError(_("There was a parsing error in the configuration file."))
            exit(errorExitCode)

def getSettings(configuration_file_path, section_name, setting_name=None):
    """
    Gets the settings and their values from the conf file.
    If only section_name is specified returns all settings in that section.
    Args:
        configuration_file_path(str): The path of the mssql-conf configuration file
        section_name(str): The name of the section a setting belongs to
        setting_name(str): Optional the setting name whose value to get

    Returns:
        A dictionary of setting_name : setting_value,
        Empty dictionary if cannot find the section or the setting
    """
    config = ConfigParser()
    results = {}
    readConfigFromFile(config, configuration_file_path)

    try:
        if not setting_name:
            for section in config.sections():
                if section_name.lower() == section.lower():
                    for setting in config.items(section):
                        results[setting[0]] = setting[1]
        else:
            results[setting_name] = config.get(section_name, setting_name)

        return results

    except Exception:
        return {}

def listSupportedSettings(supportedSettingsList):
    """List supported settings
    """

    maxLength = 0

    for setting in supportedSettingsList:
        settingLength = len("%s.%s" % (setting.section, setting.name))
        if settingLength > maxLength:
            maxLength = settingLength

    def getSettingSortKey(item):
        return "%s.%s" % (item.section, item.name)

    formatString = "%-" + str(maxLength) + "s %s"
    for setting in sorted(supportedSettingsList, key=getSettingSortKey):
        if setting.hidden == False:
            print(formatString % ("%s.%s" % (setting.section, setting.name), setting.description))

    exit(successExitCode)

def printRestartRequiredMessage():
    """Print a message telling the user that SQL Server needs to be restarted.
    """

    print(_("SQL Server needs to be restarted in order to apply this setting. Please run"))
    print(_("'systemctl restart mssql-server.service'."))

def printLaunchpadRestartRequiredMessage():
    """Print a message telling the user that SQL Server Extensibility Launchpad Daemon needs to be restarted.
    """

    print(_("SQL Server Extensibility Launchpad Daemon needs to be restarted in order to apply this setting. Please run"))
    print(_("'systemctl restart mssql-launchpadd.service'."))

def printStartSqlServerMessage():
    """Print a message telling the user to start SQL Server.
    """

    print(_("Please run 'sudo systemctl start mssql-server' to start SQL Server."))

def validateCollation(collation):
    """Validate collation

    Args:
        collation(str): Collation name
    """

    directoryOfScript = os.path.dirname(os.path.realpath(__file__))
    with open(directoryOfScript + '/collations.txt') as f:
        supportedCollationsList = [line.strip() for line in f.readlines()]
    if collation.lower() in (collation.lower() for collation in supportedCollationsList):
        return True
    else:
        printError(_("'%s' is not a supported SQL Server collation.") % collation)
        return False

def validatePasswordAndPrintIfError(password):
    """Validate a password and print an error if it's not formed correctly.

    Args:
         password(str): Password

    Returns:
         True if valid, False otherwise
    """

    passwordValidationResult = isValidPassword(password)

    if passwordValidationResult != passwordErrorCodes.SUCCESS:
        printPasswordErrorMessage(passwordValidationResult)
        return False
    else:
        return True

def getSystemAdministratorPassword(noprompt):
    """Get the system administrator password from the user via environment
    variable or interactively.

    Returns:
        UTF-8 encoded system administrator password
    """

    if checkRunningInstance():
        return None

    passwordFromEnvironmentVariable = os.environ.get(mssqlSaPasswordEnvVariable)

    # If MSSQL_SA_PASSWORD is not set, fall back to SA_PASSWORD
    #
    if passwordFromEnvironmentVariable is None:
        passwordFromEnvironmentVariable = os.environ.get(saPasswordEnvVariable)

    if (passwordFromEnvironmentVariable != None):
        if(validatePasswordAndPrintIfError(passwordFromEnvironmentVariable) == False):
            return errorExitCode
        return encodeSystemAdministratorPassword(passwordFromEnvironmentVariable)

    if not noprompt or noprompt == False:
        return encodeSystemAdministratorPassword(getSystemAdministratorPasswordInteractive())

    print(_("The MSSQL_SA_PASSWORD environment variable must be set in order to change the"))
    print(_("system administrator password."))

    return None

def getSystemAdministratorPasswordInteractive():
    """Get the system administrator password from the user interactively.

    Returns:
        System administrator password
    """

    while True:
        saPassword = getpass.getpass(_("Enter the SQL Server system administrator password: "))

        if validatePasswordAndPrintIfError(saPassword) == False:
            continue

        saPasswordConfirm = getpass.getpass(_("Confirm the SQL Server system administrator password: "))

        if (saPassword != saPasswordConfirm):
            printError(_("The passwords do not match. Please try again."))
            continue

        break

    return saPassword

def encodeSystemAdministratorPassword(password):
    """Encode system administrator password in UTF-8 format

    Args:
        noprompt(bool): Don't prompt user if True

    Returns:
        UTF-8 encoded system administrator password
    """

    # Warn the user if the system encoding cannot be detected.
    #
    if sys.stdin.encoding is None:
        print((_("Input encoding cannot be detected. SQL Server will attempt to "
                            "decode input as UTF-8")))
        encoding = 'utf-8'
    else:
        encoding = sys.stdin.encoding

    encodedPassword = password.decode(encoding).encode('utf-8')

    return encodedPassword

def validatePid(pid):
    """Validate a product key

    Args:
        pid(str): Product key

    Returns:
        Product key if valid, otherwise None
    """

    if not (
        pid.lower() == expressEdition or
        pid.lower() == evaluationEdition or
        pid.lower() == developerEdition or
        pid.lower() == webEdition or
        pid.lower() == standardEdition or
        pid.lower() == enterpriseEdition or
        pid.lower() == enterpriseCoreEdition or
        re.match("^([A-Za-z0-9]){5}-([A-Za-z0-9]){5}\-([A-Za-z0-9]){5}\-([A-Za-z0-9]){5}\-([A-Za-z0-9]){5}$", pid)
    ):
        printError(_("Invalid PID specified: %s.") % (pid))
        print("")
        return None

    return pid

def getPidFromEditionSelected(edition):
    """Gets the correct pid to pass to the engine

    Args:
        edition(string): Edition option 1-8

    Returns:
        Pid as expected by the engine

    """

    if edition == "1":
        return evaluationEdition
    elif edition == "2":
        return developerEdition
    elif edition == "3":
        return expressEdition
    elif edition == "4":
        return webEdition
    elif edition == "5":
        return standardEdition
    elif edition == "6":
        return enterpriseEdition
    elif edition == "7":
        return enterpriseCoreEdition
    elif edition == "8":
        while(True):
            productKey = raw_input(_("Enter the 25-character product key: "))
            print("")
            if validatePid(productKey):
                break
        return productKey

    else:
        print(_("Invalid option %s.") % edition)
        exit(errorExitCode)

def getPid(noprompt=False):
    """Get product key from user

    Args:
        noprompt(bool): Don't prompt user if True

    Returns:
        Product key
    """

    pidFromEnv = os.environ.get(mssqlPidEnvVariable)

    if (pidFromEnv != None):
        return validatePid(pidFromEnv)

    # If running with --noprompt and MSSQL_PID not set return developer edition
    #
    if (noprompt):
        return developerEdition

    print(_("Choose an edition of SQL Server:"))
    print("  1) Evaluation " + _("(free, no production use rights, 180-day limit)"))
    print("  2) Developer " + _("(free, no production use rights)"))
    print("  3) Express " + _("(free)"))
    print("  4) Web " + _("(PAID)"))
    print("  5) Standard " + _("(PAID)"))
    print("  6) Enterprise " + _("(PAID)") + " - " + _("CPU Core utilization restricted to 20 physical/40 hyperthreaded"))
    print("  7) Enterprise Core " + _("(PAID)") + " - " + _("CPU Core utilization up to Operating System Maximum"))
    print("  8) " + _("I bought a license through a retail sales channel and have a product key to enter."))
    print("")
    print(_("Details about editions can be found at"))
    print(getFwlinkWithLocale("2109348"))
    print("")
    print(_("Use of PAID editions of this software requires separate licensing through a"))
    print(_("Microsoft Volume Licensing program."))
    print(_("By choosing a PAID edition, you are verifying that you have the appropriate"))
    print(_("number of licenses in place to install and run this software."))
    print("")
    edition = raw_input(_("Enter your edition") + "(1-8): " )

    pid = getPidFromEditionSelected(edition)
    return validatePid(pid)

def configureSqlservrWithArguments(*args, **kwargs):
    """Configure SQL Server with arguments

    Args:
        args(str): Parameters to SQL Server
        kwargs(dict): Environment variables

    Returns:
        SQL Server exit code
    """

    args = [invokeSqlservrScript] + list(args)
    env = dict(os.environ)
    env.update(kwargs)
    env = {str(k): str(v) for k, v in env.items()}
    print(_("Configuring SQL Server..."))
    return subprocess.call(args, env=env)

def runScript(pathToScript, runAsRoot=False):
    """Runs a script (optionally as root)

    Args:
        pathToScript(str): Path to script to run
        runAsRoot(boolean): Run script as root or not

    Returns:
        Script exit code
    """

    if (runAsRoot):
        if(checkSudo() == False):
            printError(_("Elevated privileges required for this action. Please run in 'sudo' mode."))
            return (errorExitCode)
        return subprocess.call([sudo, "-EH", pathToScript])
    else:
        return subprocess.call([pathToScript])

def checkInstall():
    """Checks installation of SQL Server

    Returns:
        True if there are no problems, False otherwise
    """

    return runScript(checkInstallScript, True) == 0

def checkRunningInstance():
    """Check if an instance of SQL Server is running

    Returns:
        True if there is an instance running, False otherwise
    """

    ret = runScript(checkRunningInstanceScript)

    if (ret == 0):
        print (_("An instance of SQL Server is running. Please stop the SQL Server service"))
        print (_("using the following command"))
        print ("")
        print ("    sudo systemctl stop mssql-server")
        return True

    return False

def setupSqlServer(eulaAccepted, noprompt=False):
    """Setup and initialize SQL Server

    Args:
        eulaAccepted (boolean): Whether Eula was accepted on command line or via env variable
        noprompt (boolean): Don't prompt if True
    """

    # Make sure installation basics are OK
    #
    if not checkInstall():
        exit(errorExitCode)

    # Check if SQL Server is already running
    #
    if checkRunningInstance():
        exit(errorExitCode)

    # Get product key
    #
    pid = getPid(noprompt)

    if(noprompt == False and pid is None):
        exit(errorExitCode)

    # Check for EULA acceptance and show EULA based on edition selected
    if not checkEulaAgreement(eulaAccepted, configurationFilePath, isEvaluationEdition = (pid == evaluationEdition)):
        printError(_("EULA not accepted. Exiting."))
        exit(errorExitCode)

    # Select language and write LCID to configuration
    #
    languageSelect(noprompt)

    # Set settings from environment
    #
    if (not setSettingsFromEnvironment()):
        exit(errorExitCode)

    # Get system administrator password
    #
    encodedPassword = getSystemAdministratorPassword(noprompt)
    if (encodedPassword == errorExitCode or encodedPassword == None):
        exit(errorExitCode)

    if (pid == None):
        ret = configureSqlservrWithArguments("--setup --reset-sa-password", MSSQL_SA_PASSWORD=encodedPassword)
    else:
        ret = configureSqlservrWithArguments("--setup --reset-sa-password", MSSQL_SA_PASSWORD=encodedPassword, MSSQL_PID=pid)

    if (ret == errorExitCode):
        print(_("Initial setup of Microsoft SQL Server failed. Please consult the ERRORLOG"))
        print(_("in %s for more information.") % (getErrorLogFile(configurationFilePath)))
        exit(ret)

    # Start the SQL Server service
    #
    ret = subprocess.call(["sudo", "systemctl", "start", "mssql-server"])
    if (ret != 0):
        print (_("Attempting to start the Microsoft SQL Server service failed."))
        exit(ret)

    # Enable SQL Server to run at startup
    #
    ret = subprocess.call(["sudo", "systemctl", "enable", "mssql-server"])
    if (ret != 0):
        print (_("Attempting to enable the Microsoft SQL Server to start at boot failed."))
        exit(ret)

    print(_("Setup has completed successfully. SQL Server is now starting."))

    return None

def isMlServicesInstalled():
    """Check whether machine learning services has been installed.
    """
    if os.path.exists(launchpaddPath) and os.path.exists(mlservicesPathRoot):
        return True

    return False

def checkInstallExtensibility():
    """Checks installation of SQL Server Extensibility service

    Returns:
        True if there are no problems, False otherwise
    """

    return runScript(checkInstallExtensibilityScript, True) == 0

def setupSqlServerMlServices(eulaMlAccepted):
    """Setup and initialize SQL Server Machine Learning services (Launchpadd, R and/or python)

    Args:
        eulaMlAccepted (boolean): Whether Eula was accepted on command line or via env variable
    """

    # Make sure installation basics are OK
    #
    if not checkInstallExtensibility():
        exit(errorExitCode)

    # Check for ML EULA acceptance
    if not checkEulaMlAgreement(eulaMlAccepted, configurationFilePath):
        printError(_("EULA for machine learning services not accepted. Exiting."))
        exit(errorExitCode)

    # Start the SQL Server ML service
    #
    ret = subprocess.call(["sudo", "systemctl", "start", "mssql-launchpadd"])
    if (ret != 0):
        print (_("Attempting to start Microsoft SQL Server Machine Learning Services failed."))
        exit(ret)

    # Enable SQL Server ML service to run at startup
    #
    ret = subprocess.call(["sudo", "systemctl", "enable", "mssql-launchpadd"])
    if (ret != 0):
        print (_("Attempting to enable Microsoft SQL Server Machine Learning Services to start at boot failed."))
        exit(ret)

    print(_("Setup has completed successfully. Microsoft SQL Server Machine Learning Services is now starting."))
    print("")

    printRestartRequiredMessage()

    return None

def isValidPassword(password):
    """Determines if a given password matches SQL Server policy requirements

    Args:
        password (str): a password encoded as described by sys.stdin.encoding

    Returns:
        Integer error code. 0 for success, [1-6] otherwise. Error codes are defined
        in the passwordErrorCodes class.

    Notes:
        This function cannot print anything to stdout or it will interfere with
        the validate-password flag in mssql-conf.py
    """

    # Decode the password as whatever the input encoding is.
    # If the environmental variable is not set, then we have no way of knowing the
    # encoding of the input text. The user will have already received a warning message.
    #
    inputEncoding = sys.stdin.encoding
    if inputEncoding is None:
        inputEncoding = 'utf-8'

    # Characters that Windows treats as "special characters" for its password validation.
    #
    SYMBOL_CHARACTERS = "(`~!@#$%^&*_-+=|\\{}[]:;\"'<>,.?)/"

    # Known bug:
    # This function will return the incorrect character count for multi-byte encodings
    # For example, a smiley face emoji, U+1F601 will cause this function to allow shorter
    # passwords. However, this matches the behavior of Windows password policy validation,
    # and will thus cause no compatability issues or information loss with SQL Server.
    #
    if len(password) < 8:
        return passwordErrorCodes.TOO_SHORT
    if len(password) > 128:
        return passwordErrorCodes.TOO_LONG

    # Exit gracefully if we get a character that cannot be decoded.
    try:
        unicodePassword = password.decode(inputEncoding)
    except (UnicodeEncodeError, UnicodeDecodeError):
        return passwordErrorCodes.DECODING_ERROR

    # Exit if we cannot convert the password to UTF-8 for the PAL
    #
    try:
        testOutput = unicodePassword.encode('utf-8')
    except (UnicodeEncodeError, UnicodeDecodeError):
        return passwordErrorCodes.ENCODING_ERROR

    # Count the number of occurences of each type of character
    #
    containsUpper = 0
    containsLower = 0
    containsNumeric = 0
    containsSymbol = 0

    for unicodeChar in unicodePassword:
        # Gets the general category assigned to a given unicode character
        # If it does not fall into any of these categories (For example, Sm, which includes U+221A)
        # then it is treated as a character for length, but does not count toward any category.
        #
        # This does not perfectly match the way that Windows evaluates characters, but is a
        # proper subset of characters accepted by Windows
        #
        # If a character is a "control character" like escape, then we print an error.
        #
        char_category = unicodedata.category(unicodeChar)
        if unicodeChar in SYMBOL_CHARACTERS:
            containsSymbol = 1
        elif char_category == 'Ll':
            containsLower = 1
        elif char_category == 'Lu':
            containsUpper = 1
        elif char_category == 'Nd':
            containsNumeric = 1
        elif char_category == 'Cc':
            return passwordErrorCodes.CONTROL_CHARS

    if (containsUpper + containsLower + containsNumeric + containsSymbol) >= 3:
        return passwordErrorCodes.SUCCESS

    else:
        return passwordErrorCodes.NOT_COMPLEX

def _printTextInColor(text, bcolor):
    """_printTextInColor

    Args:
        text(str): Text to print
        bcolor(int): Color to use
    """

    if (checkColorSupported()):
        print(bcolor + text + bcolors.ENDC)
    else:
        print(text)

def initialize():
    """Initialize mssqlconfhelper
    """

    try:
        defaultMoFilePath = os.path.dirname(os.path.realpath(__file__)) + "/loc/mo/mssql-conf-en_US.mo"
        locale.setlocale(locale.LC_ALL, '')
        localeCode = locale.getlocale()[0]

        if (localeCode == None):
            moFilePath = defaultMoFilePath
        else:
            moFilePath = os.path.dirname(os.path.realpath(__file__)) + "/loc/mo/mssql-conf-" + localeCode + ".mo"
            if (os.path.isfile(moFilePath) == False):
                print ("Locale %s not supported. Using en_US." % localeCode)
                moFilePath = defaultMoFilePath
    except:
        print ("Error in localization. Using en_US.")
        moFilePath = defaultMoFilePath

    locstrings = gettext.GNUTranslations(open(moFilePath, "rb"))
    locstrings.install()

def getErrorLogFile(configFilePath = configurationFilePath):
    """Get error log file
    """
    config = ConfigParser()
    readConfigFromFile(config, configFilePath)

    try:
        errorlog = config.get("filelocation", "errorlogfile")
    except:
        errorlog = '/var/opt/mssql/log'

    return errorlog

class adconfig():
    def __init__(self, keytab, realm):
        if checkLoggedIn() != successExitCode:
            print(_("Error: Root must be logged in as an AD user to run this command. Please run 'sudo kinit <user>"))
            exit(errorExitCode)

        # Check if mssql.conf has an entry for kerberos keytab
        #
        config = ConfigParser()
        readConfigFromFile(config, configurationFilePath)

        keytabSetting = getSettings(configurationFilePath,
                                    "network",
                                    "kerberoskeytabfile")

        if len(keytabSetting) == 0:
            print(_("Warning: No keytab specified in mssql.conf. Please run 'mssql-conf set network.kerberoskeytabfile <path>'. /etc/krb5.keytab will be used by default"))

        if keytab == "":
            if keytabSetting != {}:
                keytab = keytabSetting["kerberoskeytabfile"]
            else:
                keytab = "/etc/krb5.keytab"

        # Check if SQL Server is running on a non-default port
        #
        portSetting = getSettings(configurationFilePath,
                                  "network",
                                  "tcpport")

        if len(portSetting) > 0:
            self.port = portSetting["tcpport"]
        else:
            self.port = "1433"

        # Initialize all member variables
        #
        self.defaultRealm = getDefaultRealm()
        self.keytab = keytab
        self.realm = realm

        self.rdns = ""
        self.spnDictionary = {}
        self.privilegedAccount = ""
        self.privilegedKvno = ""

    def validate(self):
        """ This function checks if the system is configured for AD
            authentication. If so, it returns successExitCode.
        """
        # Check a default realm is set
        #
        if self.defaultRealm == "":
            print(_("Error: No deafult realm found in /etc/krb5.conf"))
            exit(errorExitCode)

        # Check we can contact the realm
        #
        shortDefaultRealm = self.defaultRealm.split('.')[0]

        if self.checkHost(shortDefaultRealm) != successExitCode:
            print(_("Error: Cannot contact default realm '%s'") % (shortDefaultRealm))
            exit(errorExitCode)

        if self.checkHost(self.defaultRealm) != successExitCode:
            print(_("Error: Cannot contact default realm '%s'") % (self.defaultRealm))
            exit(errorExitCode)

        if self.realm != "":
            shortRealm = self.realm.split('.')[0]
            if self.checkHost(shortRealm) != successExitCode:
                print(_("Error: Cannot contact realm '%s'") % (shortRealm))
                exit(errorExitCode)

            if self.checkHost(self.realm) != successExitCode:
                print(_("Error: Cannot contact realm '%s'") % (self.realm))
                exit(errorExitCode)

        # Check RDNS is setup correctly
        #
        if self.checkRDNS() != successExitCode:
            print(_("Warning: RDNS could not resolve this host. This is not a fatal error, but should be fixed by updating your RDNS records"))

        # Check keytab is readable
        #
        if self.checkKeytabReadable() != successExitCode:
            print(_("Error: Keytab %s is not readable or does not exist") % (self.keytab))
            exit(errorExitCode)

        # Check the SPN entries are present in the keytab
        #
        if self.checkSPNs() != successExitCode:
            print(_("Error: SPN entries are not present in %s or have the wrong KVNO") % (self.keytab))
            exit(errorExitCode)

        # Check the privileged account (e.g. UPN) is present in the keytab
        #
        if self.checkPrivilegedEntry() != successExitCode:
            print(_("Error: %s entries are not present in %s or have the wrong KVNO") % (self.privilegedAccount, self.keytab))
            exit(errorExitCode)

        # Print configuration so user can confirm the values are what they
        # expected
        #
        self.toString()

        exit(successExitCode)

    def checkHost(self, host):
        """ Check if a host can be contacted by pinging said host several times
        """

        ret = successExitCode
        pingCommand = "ping -c 3 %s" % (host)

        try:
            process = subprocess.Popen(pingCommand.split(), stdout=open(os.devnull, 'w'), stderr=subprocess.PIPE)
            output, error = process.communicate()
            ret = successExitCode
        except OSError:
            print _("OS error encountered when checking if host %s can be contacted") % (host)

        if process.returncode != 0:
            ret = errorExitCode

        return ret

    def checkRDNS(self):
        """ Checks if the reverse DNS entries for this machine are configured
        """

        # Get our IP address
        #
        fqdn = socket.getfqdn()
        ip = socket.gethostbyname(fqdn)

        # Perform reverse DNS lookup on IP
        #
        try:
            hostname = socket.gethostbyaddr(ip)
        except socket.herror:
            return errorExitCode

        # Check the rDNS result is in the domain we think we are part of
        #
        self.rdns = hostname[0]

        if self.realm != "":
            matchRealm = self.realm
        else:
            matchRealm = self.defaultRealm

        matchObject = re.match(".*%s" % (matchRealm), self.rdns)

        if matchObject != None and matchObject.group() == self.rdns:
            return successExitCode
        else:
            return errorExitCode

    def checkKeytabReadable(self):
        """ Check if the keytab is readable by the mssql user
        """

        ret = errorExitCode

        try:
            readKeytabCommand = _("sudo -u mssql cat %s") % (self.keytab)
            process = subprocess.Popen(readKeytabCommand.split(), stdout=open(os.devnull, 'w'), stderr=open(os.devnull, 'w'))
            process.communicate()

            if process.returncode == 0:
                ret = successExitCode
        except OSError:
            print _("OS error encountered when checking if %s is readable") % (self.keytab)

        return ret

    def checkKeytabEntry(self, entryName, isPrivilegedAccount = False):
        """ Check if the specified service / account is in the keytab and has the latest KVNO
        """

        # Get the keytab entries
        #
        readKeytabCommand = "klist -kt %s" % (self.keytab)
        ret = errorExitCode

        try:
            readKeytabProcess = subprocess.Popen(readKeytabCommand.split(), stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
            keytabContents, error = readKeytabProcess.communicate()
            ret = readKeytabProcess.returncode
        except OSError:
            print _("OS error encountered when reading the keytab %s") % (self.keytab)

        kvno = 0

        if ret == 0:
            # Get the kvno for the requested entry
            #
            if isPrivilegedAccount:
                kvno = self.privilegedKvno
                ret = 0
            else:
                getKvnoCommand = "kvno %s" % (entryName)

                try:
                    kvnoProcess = subprocess.Popen(getKvnoCommand.split(), stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
                    kvno, error = kvnoProcess.communicate()
                    ret = kvnoProcess.returncode
                except OSError:
                    print _("OS error encountered when finding the KVNO for %s") % (entryName)
                    ret = errorExitCode

                if ret == 0:
                    kvno = kvno.split('=')

                    if len(kvno) > 1:
                        # Use regex to search for the account in the keytab. The
                        # kvno can be either the current kvno or 255
                        #
                        kvno = kvno[1].strip()
                    else:
                        ret = errorExitCode

            if ret == 0:
                ret = errorExitCode
                keytabSplit = keytabContents.split("\n")
                keytabSplitLen = len(keytabSplit)

                for keytabLine in keytabSplit:
                    keytabLine = keytabLine.strip()
                    
                    # Each line in the keytab will be stored as (kvno,
                    # Timestamp day, timestamp time, principal)
                    #
                    keytabEntry = keytabLine.split(' ')

                    if len(keytabEntry) == 4:
                        entryKvno = keytabEntry[0]
                        entryPrincipal = keytabEntry[3]

                        # Check the kvno is either the kvno we expect or 255
                        #
                        if entryKvno == kvno or entryKvno == 255:
                            # attempt to match the principal
                            #
                            entryMatchObject = re.match("%s@.*" % re.escape(entryName), entryPrincipal, re.IGNORECASE)

                            if entryMatchObject != None:
                                ret = successExitCode
                                kvno = entryKvno
                                break

        else:
            ret = errorExitCode

        return kvno, ret

    def checkSPNs(self):
        """ Check if the required SPNs are in the keytab
        """

        hostname = self.rdns.split('.')[0]
        realm = self.defaultRealm

        if self.realm != "":
            realm = self.realm

        # Generate all possible SPNs
        #
        validSPNs = []
        validSPNs.append("MSSQLSvc/%s.%s:%s" % (hostname, realm, self.port))
        validSPNs.append("MSSQLSvc/%s.%s" % (hostname, realm))
        validSPNs.append("MSSQLSvc/%s:%s" % (hostname, self.port))
        validSPNs.append("MSSQLSvc/%s" % (hostname))

        # Check if SPNs are present
        #
        for spn in validSPNs:
            kvno, ret = self.checkKeytabEntry(spn)

            if ret == successExitCode:
                self.spnDictionary[spn] = kvno

        # Privileged account should have the same KVNO as the SPNs assigned to it
        #
        self.privilegedKvno = kvno

        # Only fail if we found no valid SPNs
        #
        ret = successExitCode

        if len(self.spnDictionary) == 0:
            ret = errorExitCode

        # Print a warning if not all SPNs are present
        #
        if len(self.spnDictionary) < len(validSPNs) and len(self.spnDictionary) > 0:
            print(_("Warning: Some but not all SPNs are present"))

        return ret

    def checkPrivilegedEntry(self):
        """ Check if the privileged entry (i.e. UPN) is present in the keytab
        """

        # Get the privileged account from mssql.conf
        #
        config = ConfigParser()
        readConfigFromFile(config, configurationFilePath)

        privilegedAccountSetting = getSettings(configurationFilePath,
                                               "network",
                                               "privilegedadaccount")

        # If no privileged account specified, default to the UPN
        #
        if privilegedAccountSetting != {}:
            privilegedAccount = privilegedAccountSetting["privilegedadaccount"]
        else:
            privilegedAccount = getUPN()

        # Check the privileged account is in the keytab
        #
        kvno, ret = self.checkKeytabEntry(privilegedAccount, True)

        self.privilegedAccount = privilegedAccount

        return ret

    def toString(self):
        """ Print stored configuration attributes
        """

        print(_("\nDetected Configuration:"))
        print(_("Default Realm: %s") % (self.defaultRealm))

        if self.realm != "":
            print(_("Realm: %s") % (self.realm))

        print(_("Keytab: %s") % (self.keytab))
        print(_("Reverse DNS Result: %s") % (self.rdns))
        print(_("SQL Server Port: %s") % (self.port))
        print(_("Detected SPNs (SPN, KVNO):"))

        for spn in self.spnDictionary:
            print("\t(%s, %s)" % (spn, self.spnDictionary[spn]))

        print(_("Privileged Account (Name, KVNO): \n\t(%s, %s)") % (self.privilegedAccount, self.privilegedKvno))

def getUPN():
    """ Generate the SPN from the hostname
    """

    fqdn = socket.getfqdn()
    hostname = fqdn.split('.')[0]
    hostname = hostname.strip()
    hostname += "$"
    return hostname

def checkLoggedIn():
    """ Checked if the user has kinit'ed
    """

    ret = errorExitCode

    try:
        klistCommand = "klist"
        process = subprocess.Popen(klistCommand.split(), stdout=open(os.devnull, 'w'), stderr=open(os.devnull, 'w'))
        process.communicate()
        ret = process.returncode
    except OSError:
        print _("OS error encountered when checking if root is logged into active directory")

    return ret

def getDefaultRealm():
    """ Find the default realm specified in /etc/krb5.conf. Note, this
        currently does not support the "include" keyword supported by krb5.conf
    """

    # Get the first default realm entry in /etc/krb5.conf that isn't commented out
    #
    defaultRealmSplit = []

    try:
        getDefaultRealmCommand = "grep -i -m 1 ^[^#]*default_realm /etc/krb5.conf"
        process = subprocess.Popen(getDefaultRealmCommand.split(), stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        defaultRealm, ret = process.communicate()
        defaultRealmSplit = defaultRealm.split('=')
    except OSError:
        print _("OS error encountered when searching /etc/krb5.conf for the default domain")

    # Get the value if present
    #
    if len(defaultRealmSplit) > 1:
        defaultRealm = defaultRealmSplit[1]
    else:
        defaultRealm = ""

    return defaultRealm.strip()

def setupADKeytab(keytab, user):
    """ Entry point for auto-generating the AD keytab for SQL Server
    """

    if checkLoggedIn() != successExitCode:
        print(_("Error: Root must be logged in as an AD user to run this command. Please run 'sudo kinit <user>"))
        exit(errorExitCode)

    # Get the hostname
    #
    hostname = socket.gethostname().strip()
    hostnameSplit = hostname.find('.')

    if hostnameSplit != -1:
        hostname = hostname[:hostnameSplit]

    # Get the domain
    #
    domain = getDefaultRealm()

    if domain == "":
        print(_("Error: Default realm must be set in /etc/krb5.conf"))
        exit(errorExitCode)

    domain = domain.lower()

    # Get the password to use for inserting keytab entries
    #
    privilegedPassword = ""
    spnPassword = ""

    if user != "":
        privilegedPassword = getpass.getpass(_("%s@%s's password: ") % (user, domain))
        spnPassword = privilegedPassword
    else:
        spnPassword = getpass.getpass(_("SPNs' password: "))

    # Get all potential SPNs
    #
    spnList = getSPNs(hostname, domain)

    # Get KVNO of all SPNs
    #
    spnsWithKvno = getKVNOsForSPNs(spnList)
    
    if len(spnsWithKvno) == 0:
        print(_("Error: No registered SPNs found. Please register SPNs then re-run this script"))
        exit(errorExitCode)

    # Add SPNs to keytab
    #
    for spn in spnsWithKvno:
        addEntryToKeytab(spn, spnsWithKvno[spn], spnPassword, keytab)

        # KVNO is the same for the privileged AD user and the SPNs it owns
        #
        privilegedUserKvno = spnsWithKvno[spn]

    # Add privileged account to keytab
    #
    if user != "":
        addEntryToKeytab(user, privilegedUserKvno, privilegedPassword, keytab)
    else:
        addUPNToKeytab(keytab)

    # Set owner of keytab
    #
    try:
        mssqlUID = pwd.getpwnam("mssql")
        os.chown(keytab, mssqlUID.pw_uid, mssqlUID.pw_gid)
    except KeyError:
        print(_("Warning: User 'mssql' does not exist. Auto-generated keytab "
                "still owned by root. Please change ownership to the user SQL "
                "Server runs as (chown mssql:mssql %s") % (keytab))

    # Set keytab path in mssql.conf
    #
    if len(keytab) > 0 and keytab[0] != '/':
        cwd = os.getcwd()

        if cwd[-1] != '/':
            cwd += '/'

        setSettingByName("network", "kerberoskeytabfile", cwd + keytab)
    else:
        setSettingByName("network", "kerberoskeytabfile", keytab)

    # Set privileged user in mssql.conf (and clear it if no privileged user)
    #
    setSettingByName("network", "privilegedadaccount", user)

    print(_("SQL Server needs to be restarted in order to addopt the new AD configuration. Please run"))
    print(_("'systemctl restart mssql-server.service'."))

def getSPNs(hostname, domain):
    """ Generate all possible SPNs
    """

    spnList = []

    # Get the port
    #
    port = spnDefaultPort
    portSetting = getSettings(configurationFilePath,
                              "network",
                              "tcpport")

    if len(portSetting) > 0:
        port = portSetting["tcpport"]

    # Generate all potential SPNs
    #
    spnList.append("%s/%s.%s:%s" % (spnPrefix, hostname, domain, port))
    spnList.append("%s/%s.%s" % (spnPrefix, hostname, domain))
    spnList.append("%s/%s:%s" % (spnPrefix, hostname, port))
    spnList.append("%s/%s" % (spnPrefix, hostname))

    return spnList

def getKVNO(principal):
    """ Retrieve the KVNO for a principal
    """

    kvno = -1
    kvnoCommand = kvnoCommandFormat % (principal)

    try:
        process = subprocess.Popen(kvnoCommand.split(), stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        kvnoRes, ret = process.communicate()

        if kvnoRes.find('=') != -1:
            kvno = kvnoRes.split('=')[1].strip()
    except OSError:
        print _("OS error when looking up KVNO for %s") % (principal)

    return kvno

def getKVNOsForSPNs(spnList):
    """ Get the KVNO for all SPN inputs
    """

    spnKvnoMap = {}

    for spn in spnList:
        kvno = getKVNO(spn)

        if kvno > 0:
            spnKvnoMap[spn] = kvno

    return spnKvnoMap

def addEntryToKeytab(entry, kvno, password, keytab):
    """ Add an entry to the keytab for all supported encryption types
    """

    if password != "":
        for encType in keytabEncryptionTypes:
            if not isEntryInKeytab(entry, kvno, encType, keytab):
                try:
                    addEntryCommand = addKeytabEntryCommandFormat % (entry, kvno, encType, password, keytab)
                    process = subprocess.Popen(addEntryCommand, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                    res, ret = process.communicate()
                except OSError:
                    print _("OS error when adding (%s, %s) to keytab %s") % (entry, kvno, keytab)
    else:
        # No password given, attempt to copy from /etc/krb5.ketyab
        #
        copyEntryFromKeytabToKeytab(entry, kvno, rootKeytab, keytab)

def addUPNToKeytab(keytab):
    """ Copy the machine account credentials from /etc/krb5.keytab to the
        output keytab
    """

    upn = getUPN()
    kvno = getKVNO(upn)

    if kvno > 0:
        copyEntryFromKeytabToKeytab(upn, kvno, rootKeytab, keytab)
    else:
        print(_("Error: Machine account credential not found in Active "
                "Directory. Ensure the machine is domain joined or supply "
                "a privileged AD account for SQL Server to authenticate as"))
        exit(errorExitCode)

def copyEntryFromKeytabToKeytab(principal, kvno, srcKeytab, destKeytab):
    """ Copy all occurrences of 'principal' with the right KVNO from the source
        keytab to the destination keytab
    """

    listKeytabContentsCommand = listKeytabContentsCommandFormat % (srcKeytab)

    try:
        process = subprocess.Popen(listKeytabContentsCommand.split(), stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        keytabContents, ret = process.communicate()

        # Get each line in the keytab, except the first three (keytab name, column
        #  names and a horizontal line)
        #
        keytabLines = keytabContents.split('\n')[3:]
        slot = 1
        slotsToDelete = []

        # Find all entries which will not be coppied
        #
        for line in keytabLines:
            lineContents = line.strip().split(' ')

            # skip empty lines
            #
            if len(lineContents) >= 5:
                # Get entrie's KVNO, name, and encryption type
                #
                entryKvno = lineContents[0]
                entryPrincipal = lineContents[3].split('@')[0]
                entryEncType = lineContents[4].strip('()')

                # If don't want to copy the slot, add it to the list
                #
                if (entryPrincipal.lower() != principal.lower()) or (entryKvno != kvno) or (isEntryInKeytab(entryPrincipal, entryKvno, entryEncType, destKeytab)):
                    slotsToDelete.append(slot)

            slot += 1

        if len(slotsToDelete) >= slot - 1:
            print(_("Error: No machine credentials in /etc/krb5.keytab to copy"))
            exit(errorExitCode)

        # Create command to read source keytab into memory, delete all irrelevant
        # entires, and write to destination keytab
        #
        slotsToDelete = reversed(slotsToDelete)
        copyKeytabCommand = copyKeytabEntriesPrefixCommandFormat % (srcKeytab)

        for slot in slotsToDelete:
            copyKeytabCommand += deleteKeytabEntryCommandFormat % (slot)

        copyKeytabCommand += copyKeytabEntriesSuffixCommandFormat % (destKeytab)

        # Execute command to copy values
        #
        process = subprocess.Popen(copyKeytabCommand, stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'), shell=True)
        process.communicate()
    except OSError:
        print _("OS error when copying keytab entry (%s, %s) from %s to %s") % (principal, kvno, srcKeytab, destKeytab)

def isEntryInKeytab(entry, kvno, encType, keytab):
    """ Set a value in mssql.conf
    """

    res = False

    # Get all the entries in the keytab
    #
    listKeytabContentsCommand = listKeytabContentsCommandFormat % (keytab)

    try:
        process = subprocess.Popen(listKeytabContentsCommand.split(), stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        keytabContents, ret = process.communicate()

        # Get each line in the keytab, except the first three (keytab name, column
        #  names and a horizontal line)
        #
        keytabLines = keytabContents.split('\n')[3:]

        # Find all entries which will not be coppied
        #
        for line in keytabLines:
            lineContents = line.strip().split(' ')

            # skip empty lines
            #
            if len(lineContents) >= 5:
                # Get entrie's KVNO, name, and enctyprion type
                #
                entryKvno = lineContents[0]
                entryPrincipal = lineContents[3].split('@')[0]
                entryEncType = lineContents[4].strip('()')

                # If don't want to copy the slot, add it to the list
                #
                if (entryPrincipal.lower() == entry.lower()) and (entryKvno == kvno) and (entryEncType.lower() == encType.lower()):
                    res = True
                    break
    except OSError:
        print _("OS error when looking up keytab entry (%s, %s, %s) in %s") % (entry, kvno, encType, keytab)

    return 